/*
    control++ : System configuration tool
    Copyright (C) 2017-2024 Alexey Appolonov

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "controller.h"
#include <algorithm>

//-----------------------------------------------------------------------------

TController::TController(void){

    keyword = "macro";

    // Controlled units of the Controller
    for(int i = 0; i < UNITS_Q; i++){
        switch(i){
            case U_ULIMITS:
                units[i] = new TULimits();
                break;
            case U_PERMS:
                units[i] = new TPermissions();
                break;
            case U_SCRIPTS:
                units[i] = new TStarter();
                break;
            default:
                units[i] = NULL;
                break;
            }
        }

    printer.Silent(true);

    // Read configuration file of specified type
    auto read_conf_file = [&](const string &type){
        const auto path = \
           EnvPath(type == "main" ? MAIN_CONF_FILE : INTERNAL_CONF_FILE);
        printer.LineBegin("Reading " + type + " conf file \"" + path + "\"");
        return NewConfFile(path);
        };

    // Controller has two config files - one for user and one for internal
    // operations
    bool flag = false;
    if(read_conf_file("main")){
        if(read_conf_file("internal")){
            flag = true;
            }
        }

    // If one of the conf files cannot be read
    if(!flag){
        printer.Silent(false);
        printer.Err();
        exit(1);
        }
    else{
        printer.Success();
        }

    printer.Silent(false);
    }

//-----------------------------------------------------------------------------

TController::~TController(void){

    for(auto &u : units){
        if(u){
            delete u;
            }
        }
    }

//-----------------------------------------------------------------------------
// Find unit by it's keyword and return the pointer

TUnit* TController::GetUnit(const string &unit_name){

    for(auto &u : units){
        if(unit_name == u->KeyWord()){
            return u;
            }
        }

    return NULL;
    }

//-----------------------------------------------------------------------------
// Display content of configuration file

bool TController::PrintConfFile(void){

    printer.LineEnd("Configuration file path:\n" +
        EnvPath(MAIN_CONF_FILE) + "\n");

    if(!configs[EXTERNAL_CONF].ReadFile(EnvPath(MAIN_CONF_FILE))){
        printer.Err("Can't read configuration file");
        return false;
        }

    configs[EXTERNAL_CONF].PrintFile();
    return true;
    }

//-----------------------------------------------------------------------------
// Display current mode for all units

void TController::PrintCurrentModesOfTheUnits(void){

    PrintCurrentMode(MACRO);

    if(!current_mode_stated_in_iconf.empty()){
        printer.IndentAdd();
        if(units_of_the_mode.empty()){
            for(const auto &unit : units){
                unit->PrintCurrentMode(MACRO, SHORT_FMT);
                }
            }
        else{
            for(const auto &unit : units_of_the_mode){
                unit->PrintCurrentMode(MACRO, SHORT_FMT);
                }
            }
        printer.IndentClear();
        }

    for(const auto &unit : units){
        unit->PrintCurrentMode(MICRO);
        }
    }

//-----------------------------------------------------------------------------
// Apply configuration file DOM to the unit

bool TController::ApplyConfDOM(void){

    // Reading current main mode from internal conf
    vector<string> internal_conf_sectors = configs[INTERNAL_CONF].AllSecNames();
    if(!internal_conf_sectors.empty() &&
            internal_conf_sectors.back() != MICRO_MODES_SECNAME){
        current_mode_stated_in_iconf = internal_conf_sectors.back();
        }
    // Empty internal conf file means that there is no mode set
    else{
        current_mode_stated_in_iconf = "";
        }

    // Finding available sub modes of all controled units
    for(auto &unit : units){
        if(!unit->FindoutAvailableModes()){
            return false;
            }
        }

    // Checking that there are some main modes of the control++
    if(!FindoutAvailableModes()){
        return false;
        }

    // Writing current main mode
    if(!WriteCurrentMode(current_mode_stated_in_iconf)){
        // If stated mode is not available then current mode is 'empty' mode
        if(!WriteCurrentMode("")){
            printer.Err("Can't write current mode");
            return false;
            }
        printer.LineEnd("Current mode is not available");
        return true;
        }

    string current_sub_mode_iconf;
    string current_sub_mode_econf;
    string current_micro_mode;

    // Writing current sub modes for all controled units
    for(auto &unit : units){

        configs[INTERNAL_CONF].ReadDOMParam(
            current_sub_mode_iconf, CurrentModeName(), unit->KeyWord()
            );
        configs[EXTERNAL_CONF].ReadDOMParam(
            current_sub_mode_econf, CurrentModeName(), unit->KeyWord()
            );
        configs[INTERNAL_CONF].ReadDOMParam(
            current_micro_mode, MICRO_MODES_SECNAME, unit->KeyWord()
            );

        unit->WriteCurrentModeStatedInIConf(current_sub_mode_iconf);

        const bool macro_mode_ok = \
            unit->WriteCurrentMode(current_sub_mode_econf);
        const bool micro_mode_ok = !macro_mode_ok ? false :
            unit->WriteCurrentMode(current_micro_mode, MICRO);

        if(!macro_mode_ok || !micro_mode_ok){
            string err_msg = "Can't initialize " + unit->KeyWord() + " unit "
                "(failed to write " + (macro_mode_ok ? "micro" : "macro") + " "
                "mode)";
            printer.Err(err_msg);
            return false;
            }
        }

    return true;
    }

//-----------------------------------------------------------------------------
// Form vector of names of the available modes

bool TController::FindoutAvailableModes(void){

    // Getting names of the main modes
    vector<string> main_modes = configs[EXTERNAL_CONF].AllSecNames();

    TConfSec main_mode_desc;
    map<string, map<string, string>> submodes;

    // Checking availability of sub modes of each main mode
    for(auto mm = (main_modes.end() - 1); mm != (main_modes.begin() - 1); mm--){

        bool corrupted = false;
        configs[EXTERNAL_CONF].GetDOMSec(*mm, main_mode_desc);

        const auto &main_mode_name = main_mode_desc.Name();

        for(auto &assgn : main_mode_desc.Assgn()){

            if(assgn.size() < 2){
                printer.Warn("'" + main_mode_name + "' mode is corrupted -"
                    "wrong assignment in '" + EnvPath(MAIN_CONF_FILE) + "'"
                    + (assgn.empty() ? "" : ", see '" + assgn.front() + "'"));
                corrupted = true;
                break;
                }

            const string &unit_name = assgn.at(0);
            const string &submode_name = assgn.at(1);

            // Validate mode of the controlled unit
            for(auto &unit : units){

                const auto &unit_keyword = unit->KeyWord();

                if(unit_keyword != unit_name){
                    continue;
                    }

                if(!unit->Valid(submode_name)){
                    printer.Warn("'" + main_mode_name + "' mode is corrupted - "
                        "'" + unit_keyword + "' unit has no available mode "
                        "'" + submode_name + "'");
                    corrupted = true;
                    break;
                    }

                submodes[*mm][unit_name] = submode_name;
                }

            if(corrupted){
                break;
                }
            }

        if(corrupted){
            main_modes.pop_back();
            continue;
            }
        }
    // - (Checking availability of sub modes of each main mode)

    if(main_modes.empty()){
        printer.Err("There are no available modes");
        return false;
        }

    WriteAvailableModes(main_modes, submodes);

    return true;
    }

//-----------------------------------------------------------------------------
// Get previous mode of the unit

string TController::PrevSubMode(const string &unit_name){

    TConfSec s;
    string prev_mode;

    if(run_mode.Flag(M_MICRO_MODESETTING)){
        configs[INTERNAL_CONF].GetDOMSec(MICRO_MODES_SECNAME, s);
        }
    else{
        configs[INTERNAL_CONF].GetLastDOMSec(s);
        }

    configs[INTERNAL_CONF].ReadDOMParam(prev_mode, s.Name(), unit_name);

    return prev_mode;
    }

//-----------------------------------------------------------------------------
// Set given mode

bool TController::SetMode(const string &mode_name, const string &unit_name){

    // If name of the unit is given than micro mode setting would be performed
    TUnit* particular_unit_to_set = NULL;
    if(!unit_name.empty()){
        particular_unit_to_set = GetUnit(unit_name);
        run_mode.WriteFlag(M_MICRO_MODESETTING, true);
        if(!particular_unit_to_set){
            printer.Err("There is no '" + unit_name + "' unit");
            return false;
            }
        }
    else{
        run_mode.WriteFlag(M_MICRO_MODESETTING, false);
        if(mode_name == MICRO_MODES_SECNAME){
            printer.Err("Name " MICRO_MODES_SECNAME " is reserved name, "
                "macro mode cannot have that name");
            return false;
            }
        }

    const bool reset = (mode_name == RESET_MODE);

    // Reporting about the current operation
    if(!reset){
        const string msg = "Setting " +
            (particular_unit_to_set ? (unit_name + " ") : "") +
            "'" + mode_name + "' mode:";
        printer.LineEnd(msg);
        }
    else{
        if((!particular_unit_to_set && current_mode_stated_in_iconf.empty()) ||
                (particular_unit_to_set && !CurrentMode())){
            printer.LineEnd("Current mode is unknown, "
                "restoring will not be performed");
            return true;
            }
        printer.LineEnd("Restoring initial state:");
        }

    // Validating the mode
    if(!mode_name.empty()){
        const bool valid_mode = particular_unit_to_set ?
            particular_unit_to_set->Valid(mode_name) :
            Valid(mode_name);
        if(!valid_mode){
            printer.Err("Unknown mode");
            return false;
            }
        }
    else{
        printer.Err("Empty mode name");
        return false;
        }

    vector<string> order;
    vector<string> new_submodes;

    // Resetting the mode or resetting and setting the new one
    if(!reset){
        // Dropping current mode before setting up the new one
        if(!SetMode(RESET_MODE, unit_name)){
            return false;
            }
        // Getting order of the units from the external conf file
        if(!particular_unit_to_set){
            order = configs[EXTERNAL_CONF].AllVarNamesOfSec(mode_name);
            }
        }
    else{
        // Getting order of the units from the internal conf file
        // and reversing it (resetting the units in FIFO order)
        if(!particular_unit_to_set){
            order = configs[INTERNAL_CONF].AllVarNamesOfSec();
            reverse(order.begin(), order.end());
            }
        }

    // There is only one unit to handle when performing micro mode setting
    units_of_the_mode.clear();
    if(particular_unit_to_set){
        units_of_the_mode.push_back(particular_unit_to_set);
        new_submodes.push_back(mode_name);
        }
    else{
        // Selecting involved units
        for(auto &o : order){
            for(auto &u : units){
                if(o == u->KeyWord()){
                    units_of_the_mode.push_back(u);
                    }
                }
            }

        // Getting new sub-mode names according to new main-mode
        // (there is no need to do this for mode reset)
        if(!reset){
            new_submodes.resize(units_of_the_mode.size());
            for(int i = 0; i < (int)units_of_the_mode.size(); i++){

                auto &new_submode = new_submodes[i];
                auto &unit_of_the_mode = units_of_the_mode[i];
                auto &type_of_unit = unit_of_the_mode->KeyWord();

                configs[EXTERNAL_CONF].ReadDOMParam(
                    new_submode, mode_name, type_of_unit);
                }
            }
        }

    bool success;

    // Trying to set new sub-mode for the units
    for(auto u = units_of_the_mode.begin(); u != units_of_the_mode.end(); u++){

        auto &unit_name = (*u)->KeyWord();
        auto index = u - units_of_the_mode.begin();

        const string msg = (reset ?
                ("Restoring initial state ") :
                ("Setting '" + mode_name + "' mode ")) +
            "for the '" + unit_name + "' unit";
        printer.LineBegin(msg);

        if(run_mode.Flag(M_VERBOSE) || unit_name == KW_PERMS){
            printer.StartInstantFeedback();
            }

        success = reset ?
            (*u)->SetMode(mode_name, PrevSubMode(unit_name)) :
            (*u)->SetMode(new_submodes[index], "");

        if(!run_mode.Flag(M_SILENT_TERMINATION)){
            if(!success){
                const string err = reset ?
                    ("Can't restore initial state") :
                    ("Can't apply mode for the " + unit_name);
                printer.Err(err);
                }
            else{
                printer.Success();
                }
            }

        if(!success){
            // Set mode failed, reset the mode
            if(!reset){
                (*u)->SetMode(RESET_MODE, PrevSubMode(unit_name));
                }
            return false;
            }
        }

    // Saving the new mode to the internal configuration file or
    // wiping the internal configuration file completely
    const bool mode_saved = reset ?
        WipeCurrentMode(unit_name) : SaveCurrentMode(mode_name, unit_name);
    if(!mode_saved){
        printer.Err("Can't write to " + EnvPath(INTERNAL_CONF_FILE) + " "
            "configuration file");
        return false;
        }

    // Mode set OK
    if((particular_unit_to_set && !particular_unit_to_set->WriteCurrentMode(mode_name)) ||
            (!particular_unit_to_set && !WriteCurrentMode(mode_name))){
        printer.Err("Can't write current mode");
        return false;
        }

    if(!reset){
        const string msg = "Mode set OK\n"
            "Current " + (particular_unit_to_set ? unit_name : "" ) + " "
            "mode is '" + mode_name + "'";
        printer.LineEnd(msg);
        }
    else{
        printer.LineEnd("Mode reset OK");
        }

    units_of_the_mode.clear();

    return true;
    }

//-----------------------------------------------------------------------------
// Drop every DOM section but leave section that stores micro modes

void TController::ClearEveryDOMSecButMicroModeSec(void){

    // Removing all sections from DOM of the internal conf file
    // besides the first one that stores the micro modes
    if(configs[INTERNAL_CONF].DOMSize() > 0){
        TConfSec first_sec;
        configs[INTERNAL_CONF].GetDOMSec(0, first_sec);
        if(first_sec.Name() != MICRO_MODES_SECNAME){
            configs[INTERNAL_CONF].CleanDOM();
            }
        else{
            while(configs[INTERNAL_CONF].DOMSize() > 1){
                configs[INTERNAL_CONF].DropLastDOMSec();
                }
            }
        }

    // Section that store micro modes is obligatory
    if(configs[INTERNAL_CONF].DOMSize() < 1){
        TConfSec sec(MICRO_MODES_SECNAME);
        configs[INTERNAL_CONF].AddDOMSec(sec);
        }
    }

//-----------------------------------------------------------------------------
// Clear internal configuration file

bool TController::WipeCurrentMode(const string &unit_name){

    if(unit_name.empty()){
        // Removing section with macro mode
        ClearEveryDOMSecButMicroModeSec();
        }
    else{
        // Removing micro mode for the given unit
        configs[INTERNAL_CONF].ChangeDOMParam(
            "", MICRO_MODES_SECNAME, unit_name, true);
        }

    return configs[INTERNAL_CONF].RewriteAccordingToDOM(INTERNAL_CONF_HEADER);
    }

//-----------------------------------------------------------------------------
// Save mode to the internal configuration file

bool TController::SaveCurrentMode(const string &new_mode,
        const string &unit_name){

    string sec_name;
    string param_name;
    string value;

    // Macro mode setting
    if(unit_name.empty()){
        sec_name = new_mode;
        ClearEveryDOMSecButMicroModeSec();
        }
    else{
        sec_name = MICRO_MODES_SECNAME;
        param_name = unit_name;
        }

    // Write all params of the new mode to DOM of the internal conf file
    for(auto &u : units_of_the_mode){
        if(unit_name.empty()){
            param_name = u->KeyWord();
            }
        value = u->CurrentModeName();
        if(value == UNKNOWN_MODE || !configs[INTERNAL_CONF].ChangeDOMParam(
                value, sec_name, param_name, true)){
            printer.Err("Can't change DOM param " + sec_name + ":" + param_name);
            return false;
            }
        }

    // Write formed sector from configuration DOM
    // to the internal configuration file
    if(!configs[INTERNAL_CONF].RewriteAccordingToDOM(INTERNAL_CONF_HEADER)){
        printer.Err("Can't save current mode to '" +
            EnvPath(INTERNAL_CONF_FILE) + "'");
        return false;
        }

    return true;
    }

//-----------------------------------------------------------------------------
// Check controversy between internal and external config files

bool TController::ConfigsControversy(void){

    bool controversy = false;

    for(const auto &u : units){
        if(u->CurrentModeName() != u->CurrentModeStatedInIConf()){
            controversy = true;
            }
        }

    return controversy;
    }

//-----------------------------------------------------------------------------
// Check mode compliance

bool TController::CheckModeCompliance(string mode_name,
        const string &unit_name){

    // If name of the unit was given than micro-mode check would be performed
    TUnit* particular_unit_to_check = NULL;
    if(!unit_name.empty()){
        particular_unit_to_check = GetUnit(unit_name);
        if(!particular_unit_to_check){
            printer.Err("There is no '" + unit_name + "' unit");
            return false;
            }
        }

    // Empty mode name equals current mode name
    if(mode_name.empty()){
        mode_name = particular_unit_to_check ?
            particular_unit_to_check->CurrentModeName(MICRO) :
            CurrentModeName(MACRO);
        }

    // Not checking compliance for the empty (ghost) mode
    if(mode_name.empty()){
        printer.LineEnd("Mode is not specified, compliance check "
            "will not be performed");
        return true;
        }

    TUnitMode* mode = NULL;
    bool valid_mode = false;
    vector<TUnit*> units_to_check;
    if(!particular_unit_to_check){
        units_to_check.assign(units.begin(), units.end());
        mode = FindMode(mode_name);
        valid_mode = mode ? true : false;
        }
    else{
        units_to_check.push_back(particular_unit_to_check);
        valid_mode = particular_unit_to_check->Valid(mode_name);
        }

    if(!valid_mode){
        const string msg = "Mode '" +
            (!unit_name.empty() ? unit_name + ":" : "") + mode_name + "' "
            "is not available";
        printer.Err(msg);
        return false;
        }

    bool consistent = true;
    for(auto &u : units_to_check){

        // Macro mode is consistent if all of it's submodes is consistent
        if(mode){
            for(auto &submode : mode->SubModes()){
                if(submode.first == u->KeyWord() &&
                        !u->CheckModeCompliance(submode.second)){
                    consistent = false;
                    }
                }
            }
        else if(!u->CheckModeCompliance(mode_name)){
            consistent = false;
            }
        }

    const string msg = "Mode '" +
        (!unit_name.empty() ? (unit_name + ":") : "") + mode_name + "' "
        "is " + (!consistent ? "not " : "") + "consistent";
    printer.LineEnd(msg);

    return consistent;
    }

//-----------------------------------------------------------------------------
// Print available macro modes and available modes of all the units

void TController::PrintAvailableMacroAndMicroModes(void){

    map<string, map<string, string>> names_of_available_submodes;
    for(const auto &mode : available_modes){
        names_of_available_submodes[mode.Name()] = mode.SubModes();
        }

    PrintAvailableModes(names_of_available_submodes);

    for(auto &u : units){
        printer.LineBreak();
        u->PrintAvailableModes();
        }
    }

//-----------------------------------------------------------------------------
