/*
    control++ : System configuration tool
    Copyright (C) 2017-2024 Alexey Appolonov

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "main.h"
#include "controller.h"

//-----------------------------------------------------------------------------
// Display help notes, the format of each note {<note.first>, <note.second>} is
// as follows:
// <note.first>:
//     <base> <note.second>
// If <note.second> is empty, then the note is just the <note.first>

void PrintHelp(const string &base){

    vector<pair<string, string>> notes = {
        {"Set control++ macro mode", "<mode_name>"},
        {"Set control++ micro mode for the unit", "<unit_name> <mode_name>"},
        {"Restore state that held before macro mode setting", "--reset"},
        {"Restore related state that held before micro mode setting",
            "<unit_name> --reset"},
        {"Display list of available modes", "--list"},
        {"Display current macro and micro modes and check compliance "
        "of the macro mode", "--status"},
        {"Display current macro and micro modes and check compliance "
        "of micro mode of the unit", "<unit_name> --status"},
        {"Show this help notes", "--help"},
        {"Use '--plain | -p' flag to enable plain output mode", ""},
        {"Use '--force | -f' flag to force setting of the mode "
        "(disables approval dialogs)", ""},
        {"Use '--nonuniq | -n' flag to disable detection of non-unique "
        "permission modes", ""},
        {"Use '--verbose | -v' flag for verbose output", ""}
        };

    for(const auto &note : notes){
        cout << note.first << (note.second.empty() ? "" : ":") << endl;
        if(note.second.empty()){
            continue;
            }
        cout << "\t" << base << " " << note.second << endl;
        }
    }

//-----------------------------------------------------------------------------

TRunMode run_mode;
TPrinter printer;

int main(int argc, char *argv[]){

    // Converting command line arguments from C-arrays to strings
    const string exe = argv[0];
    const auto params = [argc, argv](void){
        vector<string> v;
        for(int i = 1; i < argc; i++){
            v.push_back(argv[i]);
            }
        return v;
        }();

    // Making the path of the working directory and the name of the program
    // available to every unit
    char *pwd = get_current_dir_name();
    if(!pwd){
        FatalError("Can't get name of the current directory");
        }
    run_mode.WriteExe(exe);
    run_mode.WriteWorkDir(pwd);

    string unit, mode;
    bool wrong_arg = false;

    #define F_HELP   "--help"
    #define F_DOM    "--dom"
    #define F_LIST   "--list"
    #define F_STATUS "--status"
    #define F_RESET  "--reset"

    map<string, bool> flags = {
        {F_HELP,   false},
        {F_DOM,    false},
        {F_LIST,   false},
        {F_STATUS, false},
        {F_RESET,  false}
        };

    // Set given flag
    auto set_flag = [&](const string &flag){
        // Only one flag can be set at a time
        for(const auto &kv : flags){
            if(kv.first != flag && kv.second == true){
                return false;
                }
            }
        if(flag == F_STATUS || flag == F_RESET){
            if(!unit.empty()){
                return false;
                }
            unit.swap(mode);
            }
        flags[flag] = true;
        return true;
        };

    // Get state of a specified flag
    auto get_flag = [&](const string &flag){
        const auto it = flags.find(flag);
        if(it == flags.end()){
            return false;
            }
        return it->second;
        };

    // Parsing the flags
    for(const auto &param : params){
        if(flags.find(param) != flags.end()){
            if(!set_flag(param)){
                wrong_arg = true;
                break;
                }
            }
        else if(param == "--plain" || param == "-p"){
            printer.Plain(true);
            }
        else if(param == "--force" || param == "-f"){
            run_mode.WriteForce(true);
            }
        else if(param == "--nonuniq" || param == "-n"){
            run_mode.WriteFlag(M_NON_UNIQUE_PERM, false);
            }
        else if(param == "--verbose" || param == "-v"){
            run_mode.WriteFlag(M_VERBOSE, true);
            }
        else if(param == "--running_in_build_env"){
            run_mode.WriteFlag(M_RUNNING_IN_BUILD_ENV, true);
            }
        else if(mode.empty() || unit.empty()){
            if(get_flag(F_STATUS) || get_flag(F_RESET)){
                if(!mode.empty() || !unit.empty()){
                    wrong_arg = true;
                    break;
                    }
                unit = param;
                }
            else if(mode.empty()){
                mode = param;
                }
            else{
                unit = mode;
                mode = param;
                }
            }
        else{
            wrong_arg = true;
            break;
            }
        }

    // Resetting a current mode equals setting a "reset" mode
    if(get_flag(F_RESET)){
        mode = RESET_MODE;
        }

    // Flags cannot be set twice so it safe to try to set the default values
    // after the given settings have been processed
    run_mode.WriteFlag(M_NON_UNIQUE_PERM, true);
    run_mode.WriteFlag(M_VERBOSE, false);
    run_mode.WriteFlag(M_RUNNING_IN_BUILD_ENV, false);

    // Display help notes and exit if there are no arguments,
    // or help is requested or some argument is invalid
    if(params.empty() || get_flag(F_HELP) || wrong_arg){
        PrintHelp(exe);
        return wrong_arg ? 1 : 0;
        }

    // Main control unit, all calls to the subunits addressed through it
    TController controller;

    // Configure the main unit - if we passed through here
    // than we need to configure the program
    if(!controller.Configure()){
        printer.LineEnd("control++ can not be run without configuration file");
        return 1;
        }

    // Display configuration file DOM and exit
    if(get_flag(F_DOM)){
        controller.PrintConfDOM();
        }

    // Display list of all modes and exit
    else if(get_flag(F_LIST)){
        controller.PrintAvailableMacroAndMicroModes();
        }

    // Display current mode, check its compliance and exit
    else if(get_flag(F_STATUS)){
        controller.PrintCurrentModesOfTheUnits();
        if(controller.ConfigsControversy()){
            cout << "Mode description was changed since the last mode setting"
                 << endl;
            if(run_mode.Force()){
                printer.LineEnd("Acknowledging the '--force' flag, "
                    "checking mode compliance according to the modified "
                    "mode description");
                }
            else if(!Approval("Continue to check mode compliance according "
                        "to the modified mode description [y/N]? ", "y")){
                return 1;
                }
            }
        if(!controller.CheckModeCompliance("", unit)){
            return 1;
            }
        }

    // Set given mode
    else{
        if(!controller.SetMode(mode, unit)){
            printer.LineEnd("Set mode failed");
            return 1;
            }
        }

    return 0;
    }

//-----------------------------------------------------------------------------
// Concat buildroot dir path with given path if control++ is running in a build
// environment, or return given path unchanged

const string EnvPath(const string &path){

    if(run_mode.Flag(M_RUNNING_IN_BUILD_ENV)){
        return JoinFilePaths(BUILDROOT_DIR, path);
        }

    return path;
    }

//-----------------------------------------------------------------------------
