/*
    control++ : System configuration tool
    Copyright (C) 2017-2024 Alexey Appolonov

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "perms.h"

#ifdef DEBUGING
#include "permissions_file_perm.h"
#include "permissions_file_acl.h"
#include "tick.h"
#else
#include <libcontrol++/permissions_file_perm.h>
#include <libcontrol++/permissions_file_acl.h>
#include <libcontrol++/tick.h>
#endif

#include <fstream>
#include <sstream>

//-----------------------------------------------------------------------------

TPermissions::TPermissions(void){

    const string &exe = run_mode.Exe();
    essentials.push_back({"/bin/sh", "/bin/bash", "/bin/bash3", "/bin/bash4"});
    essentials.push_back({"/bin/su"});
    essentials.push_back({"/bin/chmod"});
    essentials.push_back({
        JoinFilePaths(run_mode.WorkDir(), exe),
        JoinFilePaths("/bin/", exe),
        JoinFilePaths("/usr/bin/", exe)});
    keyword = KW_PERMS;
    }

//-----------------------------------------------------------------------------
// Push new file mode to the perms/acls vector

bool TPermissions::AddFileMode(TFileMode *&new_mode, const bool en_check){

    if(!new_mode){
        return false;
        }

    if(en_check && new_mode->Neutral()){
        delete new_mode;
        }
    else{
        if(run_mode.Flag(M_VERBOSE)){
            printer.LineAddExtra(
                "Adding mode for '" + new_mode->Path() + "' file");
            }
        PushBack(modes, new_mode, modes.size()*2, 1024);
        }

    new_mode = NULL;

    return true;
    }

//-----------------------------------------------------------------------------
// Disallow execution of every file in '/'

bool TPermissions::ModeToDisallowExeOfEverything(const string &base_dir,
        const vector<string> &except){

    bool success = false;

    if(!current_sector.ACL()){
        TFilePerm* mode = new TFilePerm();
        if(!base_dir.empty()){
            mode->WritePath(base_dir);
            }
        else{
            mode->WritePath("/");
            }
        mode->WriteOwner("*");
        mode->WriteGroup("*");
        mode->WriteMode(DISALLOW_EXE);
        success = AddDirMode(mode, except, RECURSIVE);
        if(mode){
            delete mode;
            }
        }
    else{
        TFileACL* mode = new TFileACL();
        if(!base_dir.empty()){
            mode->WritePath(base_dir);
            }
        else{
            mode->WritePath("/");
            }
        mode->WriteUser("**-");
        mode->WriteGroup("**-");
        mode->WriteMask("rw-");
        mode->WriteOther("**-");
        success = AddDirMode(mode, except, RECURSIVE);
        if(mode){
            delete mode;
            }
        }

    return success;
    }

//-----------------------------------------------------------------------------
// Use file-paths from the list file to complement vector of perms/acls;
// Every listed file will have the same mode

bool TPermissions::AddListMode(TFileMode *mode){

    vector<string> list_paths;
    const vector<string> exclude = group_of_files.BaseDirExcludedPaths();
    TFileInMem list_file;

    // Trying to read list-file and push filenames to the list_paths
    if(list_file.ReadFileToMem(mode->Path())){
        // Put list-file data string into string stream
        istringstream iss(list_file.data);
        string line;
        // For all lines of the string stream
        while(getline(iss,line)){
            // Saving lines that consist of some symbols besides whitespaces
            // and not commented out with "//" combination
            Trim(line);
            if(!line.empty() && (line.size() < 2 || line.substr(0, 2) != "//")){
                line = JoinFilePaths(group_of_files.BaseDir(), line);
                // Checking if file is not excluded and not in the excluded dir
                if(!FileIsSomewhereInThere(line, exclude)){
                    PushBack(list_paths, line, list_paths.size()*2, 128);
                    }
                }
            }
        }
    else{
        printer.LineAddErr("Can't read list file '" + mode->Path() + "'");
        return false;
        }

    // Whitelist requires additional check
    if(current_sector.WhiteList()){
        // Adding disallowing modes of everything
        vector<string> complete_exclude = exclude;
        complete_exclude.insert(complete_exclude.end(), list_paths.begin(),
            list_paths.end());
        if(!ModeToDisallowExeOfEverything(group_of_files.BaseDir(),
                complete_exclude)){
            return false;
            }
        // Forming the vector of essentials that are not in the whitelist
        // and are not excluded
        vector<string> disallowed_essentials;
        // For every set of variants
        for(auto &es : essentials){
            vector<string> filtered_es;
            // For every variant of essential ap
            for(auto &es_var : es){
                if(FileIsSomewhereInThere(es_var, group_of_files.BaseDir()) &&
                        !FileIsSomewhereInThere(es_var, complete_exclude)){
                    filtered_es.push_back(es_var);
                    }
                }
            // If every variant is not in the whitelist
            if(filtered_es.size() == es.size()){
                disallowed_essentials.insert(disallowed_essentials.end(),
                    filtered_es.begin(), filtered_es.end());
                for(auto &es_var : es){
                    printer.LineAddExtra(es_var);
                    }
                }
            }
        // Warning if some of the essential programs will be affect
        // by the whitelist mode setting
        if(!disallowed_essentials.empty()){
            printer.LineEnd("[INTERRUPTION: The following executables "
                "are not in the whitelist]");
            cout << "It can be impossible to change any permissions latter "
                    "(to reset the permission mode back, for example)" << endl;
            if(run_mode.Force()){
                printer.LineEnd("Acknowledging the '--force' flag, "
                    "the operation will be continued");
                }
            else if(!Approval("Continue anyway [y/N]? ", "y")){
                run_mode.WriteFlag(M_SILENT_TERMINATION, true);
                return false;
                }
            printer.LineRestart();
            }
        }

    // Forming vector of perms/acls using formed vector of file paths
    TFileMode* new_entry = NULL;
    bool success = true;
    for(auto &path : list_paths){
        new_entry = mode->Spawn();
        if(!new_entry){
            return false;
            }
        new_entry->WritePath(FormatFilePath(path));
        const bool is_dir = IsDir(new_entry->Path());
        // Pushing the mode if sector does not requires recursive processing
        if(!current_sector.Recursive() || !is_dir){
            // Using stated mode for dirs
            if(is_dir && !group_of_files.DirsMode().empty()){
                new_entry->CopyMode(group_of_files.DirsMode());
                }
            success = AddFileMode(new_entry);
            }
        // Processing files of the listed dir if sector requires
        // recursive processing
        else{
            success = AddDirMode(new_entry, true);
            }
        if(new_entry){
            delete new_entry;
            new_entry = NULL;
            }
        if(!success){
            goto AddListModeReturn;
            }
        }

    AddListModeReturn:
    return success;
    }

//-----------------------------------------------------------------------------
// Add new permissions for all files of the dir

bool TPermissions::AddDirMode(TFileMode *mode, const vector<string> &exclude,
        const bool recursive){

    vector<string> dirs;
    vector<string> complete_exclude;
    const vector<string> mode_data = mode->GetData();
    TFileMode *new_mode = NULL;
    TFileMode *reg_files_proto_mode = NULL;
    TFileMode *dirs_proto_mode = NULL;
    const bool neutral_dirs_mode = group_of_files.Neutral();

    // Check if file is regular file, symlink or dir
    auto filetype_check = [](unsigned char d_type){
        return (d_type == DT_REG || d_type == DT_DIR);
        };

    // Check if file is excluded
    auto exclude_check = [&](const string & file_path){
        return FileIsInThere(file_path, complete_exclude);
        };

    // Spawn new TFileMode obj for dirs
    auto spawn_dirs_mode = [&](void){
        if(!dirs_proto_mode){
            dirs_proto_mode = mode->Spawn();
            if(!group_of_files.DirsMode().empty()){
                dirs_proto_mode->CopyMode(group_of_files.DirsMode());
                }
            return dirs_proto_mode;
            }
        return dirs_proto_mode->Spawn(true);
        };

    // Spawn new TFileMode obj for regular files
    auto spawn_reg_files_mode = [&](void){
        if(!reg_files_proto_mode){
            reg_files_proto_mode = mode->Spawn();
            return reg_files_proto_mode;
            }
        return reg_files_proto_mode->Spawn(true);
        };

    // Form perms/acls for all contained files of an opened dir
    auto inside_the_dir = [&](const string& f_path, dirent *f_ent){
        // Only regular files, symlinks and dirs will be included
        if(!filetype_check(f_ent->d_type)){
            return true;
            }
        // Skipping the file if it's name presented in the exclude list
        // or in the 'excluded_paths' field of the dirs_perm
        if(exclude_check(f_path)){
            return true;
            }
        // If file is a dir
        if((f_ent->d_type == DT_DIR)){
            // If file is a './' or '../' dir
            if(DotsDir(f_ent->d_name)){
                return true;
                }
            // Not storing neutral dirs mode
            if(neutral_dirs_mode){
                dirs.push_back(f_path);
                return true;
                }
            new_mode = spawn_dirs_mode();
            }
        else{
            new_mode = spawn_reg_files_mode();
            }
        new_mode->WritePath(f_path);
        // Saving nested dirs
        if(recursive && f_ent->d_type == DT_DIR){
            dirs.push_back(f_path);
            }
        // Pushing formed permission mode
        if(!AddFileMode(new_mode)){
            return false;
            }
        return true;
        };

    // Merging given 'exclude' vector and excluded paths of 'group_of_files'
    complete_exclude.insert(complete_exclude.end(),
        group_of_files.ExcludedPaths().begin(),
        group_of_files.ExcludedPaths().end());
    for(auto &file_path : exclude){
        if(!FileIsInThere(file_path, group_of_files.ExcludedPaths())){
            complete_exclude.push_back(file_path);
            }
        }

    // Adding mode for top-level dir itself
    dirs.push_back(FormatFilePath(mode->Path()));
    if(mode->Path() != "/"){
        if(exclude_check(mode->Path())){
            // File path is excluded, nothing to do
            return true;
            }
        new_mode = spawn_dirs_mode();
        if(!AddFileMode(new_mode, IS_DIR)){
            if(new_mode){
                delete new_mode;
                }
            return false;
            }
        }

    string err;
    bool success = true;

    // While there is a dir to follow
    while(!dirs.empty()){
        // Taking path of the current dir from the top
        const string current_dir_path = DirEnding(dirs.back());
        dirs.pop_back();
        // Open current dir and form perms/acls for all contained files
        success = OpenAndPerformOnAllFiles(current_dir_path, inside_the_dir, err);
        if(new_mode){
            delete new_mode;
            new_mode = NULL;
            }
        if(!success){
            printer.LineAddErr(err);
            break;
            }
        }

    return success;
    }

//-----------------------------------------------------------------------------
// Apply current permission type

bool TPermissions::ApplyMode(const string &mode_name, const string &prev_mode){

    const string basic_err = "Can't apply " + mode_name + " permission mode";

    // Clearing the data that could exist if prev mode setting was unsuccessful
    Clear();

    // Get full path of current permissions description file and then parse it
    // and apply stated permissions (will cause call of the ApplyConfDOM func
    // and formation of the perms and acls vectors)
    if(!NewConfFile(mode_name, NOT_A_DIR) || !Configure()){
        if(mode_name == RESET_MODE && prev_mode.empty()){
            printer.LineAddWarn("There is no 'reset' file");
            return true;
            }
        if(!run_mode.Flag(M_SILENT_TERMINATION)){
            printer.Err(basic_err);
            }
        return false;
        }

    // Get current stat of the files
    if(mode_name != RESET_MODE){
        GetCurrentModes();
        }

    const bool verbose = run_mode.Flag(M_VERBOSE);

    TTick tick([](const unsigned val)->string{
        return to_string(val) + " file" + string(val == 1 ? " " : "s ") +
            "processed";
        });

    // Apply all valid entries from the permissions decription file
    string err;
    // This pointer used to pass new file mode from the mode setting function
    string *new_perm = verbose ? new string : NULL;
    for(auto &mode : modes){
        if(!mode->Access()){
            printer.LineRewriteExtra("[WARNING: Can't access '" +
                mode->Path() + "' file]");
            continue;
            }
        if(!verbose && !tick.Activated()){
            tick.Start();
            }
        if(!mode->Apply(err, new_perm)){
            if(!err.empty()){
                printer.LineAddErr(err);
                }
            printer.Err(basic_err);
            tick.Stop();
            return false;
            }
        tick.Count();
        // Adding 'ls -l' type of str if verbose output is requested
        if(verbose && new_perm){
            printer.LineAddExtra("Mode set {" + *new_perm + "}");
            }
        }
    tick.Stop();

    if(new_perm){
        delete new_perm;
        }

    ClearModes(modes);

    // Save stat of the files taken before mode setting into the "reset" file
    if(mode_name != RESET_MODE){
        return SaveStateHeldBeforeModeSetting();
        }

    return true;
    }

//-----------------------------------------------------------------------------
// Apply permissions mode stated in configuration DOM

bool TPermissions::ApplyConfDOM(void){

    TConfSec sec;
    TFileMode *mode;
    string err;
    const bool verbose = run_mode.Flag(M_VERBOSE);

    TTick tick([&](const unsigned val)->string{
        return to_string(modes.size()) + " prepared file permission mode" +
            string(modes.size() == 1 ? " " : "s ") +  "to process";
        });

    // Get next DOM section
    for(int i = 0; configs.front().GetDOMSec(i, sec); i++){

        current_sector.FormType(sec.Name());

        if(!current_sector.Valid()){
            continue;
            }

        if(!current_sector.ACL()){
            mode = new TFilePerm();
            }
        else{
            continue; // THIS LINE IS TEMPORARY
            mode = new TFileACL();
            }

        // Searching for permission/acl params in section
        for(const auto &assgn : sec.Assgn()){
            if(assgn.size() < 1){
                break;
                }
            if(assgn.size() < 2){
                continue;
                }
            // Assigning permission/acl params from DOM
            if(group_of_files.Assgn(assgn.front(), VSub(assgn,1))){
                continue;
                }
            if(mode->Assgn(assgn.front(), assgn.at(1))){
                continue;
                }
            printer.LineAddErr("Wrong format");
            delete mode;
            tick.Stop();
            return false;
            }

        // Revising extracted params
        if(!mode->Revise(current_sector, err)){
            if(!err.empty()){
                printer.LineAddExtra(err);
                }
            printer.LineAddErr("Some of the permission params are missing "
                "or not valid in '" + sec.Name() + "' sector "
                "of the '" + configs.front().FilePath() + "' file");
            delete mode;
            tick.Stop();
            return false;
            }

        // Params are OK, extend the perms vector in the way that depends
        // on the sector type
        bool success = false;
        if(!verbose && !tick.Activated()){
            tick.Start();
            }
        if(current_sector.File()){
            success = AddFileMode(mode);
            }
        else if(current_sector.Dir()){
            success = AddDirMode(mode, current_sector.Recursive());
            }
        else if(current_sector.List()){
            success = AddListMode(mode);
            }

        if(!success && !run_mode.Flag(M_SILENT_TERMINATION)){
            const string msg = "Can't form permissions for the "
                "'" + sec.Name() + "' section" +
                (mode ? " (path = '" + mode->Path() + "')" : "");
            printer.LineAddErr(msg);
            }

        if(mode){
            delete mode;
            mode = NULL;
            }

        if(!success){
            tick.Stop();
            return false;
            }
        }
    tick.Stop();

    // There could be only one perm and one not contradicting acl mode setting
    // for every particular file
    if(run_mode.Flag(M_NON_UNIQUE_PERM)){
        UniquifyModes();
        }

    return true;
    }

//-----------------------------------------------------------------------------
// Remove repetitions from formed perms and acls

void TPermissions::UniquifyModes(void){

    vector<int> indexes_to_remove;
    map<string, int> paths;
    pair<map<string, int>::iterator, bool> insert_ret;

    // Forming vector of indexes of the elements, that will be removed
    for(int i = (modes.size() - 1); i >= 0; i--){
        if(!modes.at(i)){
            continue;
            }
        insert_ret = paths.insert(pair<string, int>(modes.at(i)->Path(), i));
        const auto &inserted_el = insert_ret.first;
        const auto &newly_inserted = insert_ret.second;
        if(!newly_inserted){
            const auto &index_of_existing_mode = inserted_el->second;
            // Checking controversy
            if(Controversy(modes.at(index_of_existing_mode), modes.at(i))){
                PushBack(indexes_to_remove, i, indexes_to_remove.size()*2, 1024);
                }
            }
        }

    paths.clear();

    // If there is some repeated modes
    if(!indexes_to_remove.empty()){
        // Forming new vector without repeated modes
        vector<TFileMode*> new_modes;
        int mem_step = modes.size() - indexes_to_remove.size();
        for(int i = 0; i< (int)modes.size(); i++){
            if(!modes.at(i)){
                continue;
                }
            if(VContents(indexes_to_remove, i)){
                printer.LineAddWarn("Controversy in mode description for the '"
                + modes.at(i)->Path() + "' file, ignoring the earlier entry");
                delete modes.at(i);
                }
            else{
                PushBack(new_modes, modes.at(i), mem_step, modes.size());
                }
            }
        modes = new_modes;
        }
    }

//-----------------------------------------------------------------------------
// Get current modes of the files, stated in 'modes' vector

bool TPermissions::GetCurrentModes(void){

    TFileMode* mp;
    string err;
    bool flag = true;

    ClearModes(modes0);

    for(auto &m : modes){

        const auto &type = m->Type();
        const auto &path = m->Path();

        if(type == TYPE_FPERM){
            mp = new TFilePerm(GetPermission(path, err));
            }
        else if(type == TYPE_FACL){
            mp = new TFileACL(GetACL(path, err));
            }
        else{
            continue;
            }

        if(!mp->Complete()){
            if(!err.empty()){
                printer.LineAddErr(err);
                }
            printer.LineAddErr("Can't get mode of the '" + path + "' file");
            delete mp;
            flag = false;
            }
        else{
            PushBack(modes0, mp, modes0.size()*2, 1024);
            }
        }

    return flag;
    }

//-----------------------------------------------------------------------------
// Write current permissions into the "reset" file

bool TPermissions::SaveStateHeldBeforeModeSetting(void){

    const string path = EnvPath(STATE_DIR) + KeyWord() + "/" RESET_MODE +
        (run_mode.Flag(M_MICRO_MODESETTING) ? "_micro" : "");

    ofstream file(path);

    if(file.good()){
        file << "# Don't change this file!" << endl
             << "# This file contents permission mode of the files" << endl
             << "# that was held before the new mode was set" << endl
             << "# and it used to reset to this condition" << endl
             << endl;
        for(auto &m : modes0){
            m->Print(file);
            }
        }
    else{
        printer.LineAddErr("Can't open/create 'reset' file in " +
            EnvPath(STATE_DIR) + KeyWord());
        return false;
        }

    return true;
    }

//-----------------------------------------------------------------------------
// Check permissions compliance for all files

bool TPermissions::CheckModeCompliance(const string &mode_name){

    bool compliance = true;
    string err;
    string warn;

    printer.LineBegin("Checking '" + mode_name + "' " + KeyWord() + " "
        "mode compliance");
    printer.StartInstantFeedback();

    // Get struct that contents information about the file
    // to determine current file mode
    if(!NewConfFile(mode_name, NOT_A_DIR) || !Configure()){
        printer.Err("There is no description file");
        return false;
        }

    // Checking compliance of every file
    for(auto &m : modes){
        if(!m->CheckCompliance(err, warn)){
            compliance = false;
            if(printer.IsSilent()){
                break;
                }
            if(!err.empty()){
                printer.LineAddErr(m->Path() + " " + err);
                }
            }
        else if(!warn.empty()){
            printer.LineAddWarn(m->Path() + " " + warn);
            }
        }

    // If all permissions complied
    if(compliance){
        printer.LineEnd("[POSITIVE]");
        }
    // If some of the permissions are not complied
    else{
        printer.LineEnd("[NEGATIVE]");
        }

    return compliance;
    }

//-----------------------------------------------------------------------------
// Free memory that mode pointers pointed to

void TPermissions::ClearModes(vector<TFileMode*> &m){

    for(auto &el : m){
        if(el){
            delete el;
            }
        }

    m.clear();
    }

//-----------------------------------------------------------------------------
