/*
    control++ : System configuration tool
    Copyright (C) 2017-2024 Alexey Appolonov

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "starter.h"
#include <unistd.h>
#include <algorithm>

//-----------------------------------------------------------------------------
// Form vector of names of the available modes

void TStarter::FindoutAvailableModesSpecialPart(dirent *ent,
        vector<string> &modes){

    // Skipping files that aren't directories or that are './' or '../' dirs
    if((ent->d_type != DT_DIR) || DotsDir(ent->d_name)){
        return;
        }

    modes.push_back(ent->d_name);

    const string exe_path = DirEnding(DirOfTheUnit() + modes.back()) + "do";

    if(access(exe_path.c_str(), X_OK) == -1){
        modes.pop_back();
        }
    }

//-----------------------------------------------------------------------------
// Run script stated in configuration file

bool TStarter::RunScript(const string &mode_name, const string &action){

    string status;
    string command;
    int return_code;
    bool res = false;

    // If type of a script is valid
    if(find(actions.begin(), actions.end(), action) == actions.end()){
        status = "[ERROR: Wrong type of action]";
        goto RunScriptReturn;
        }

    // Path for a given mode
    command = ModeDescriptionPath(mode_name, IS_DIR, action == "do");
    if(command.empty()){
        status = "[ERROR: " + mode_name + " mode is corrupted - "
            "the dir does not exist]";
        goto RunScriptReturn;
        }

    // Path for a given action (script) of the mode
    command += action;
    if(access(command.c_str(), X_OK) == -1){
        // It's OK to have no 'undo' or 'test' script
        if(action == "undo" || action == "test"){
            status = "[WARNING: There is no '" + action + "' script]";
            res = true;
            goto RunScriptReturn;
            }
        else{
            status = "[ERROR: The script does not exist]";
            goto RunScriptReturn;
            }
        }

    command += " >/dev/null 2>&1";

    // Running a script
    return_code = system(command.c_str());
    if(WEXITSTATUS(return_code) == 0){
        status = (action == "test") ? "[POSITIVE]" : "[DONE]";
        res = true;
        }
    else{
        status = ((action == "test") ? "[NEGATIVE: " : "[ERROR: ") +
            ("Return code is '" + to_string(WEXITSTATUS(return_code)) + "']");
        res = false;
        }

    RunScriptReturn:

    printer.LineEnd(status);

    return res;
    }

//-----------------------------------------------------------------------------
// Run "do" or "undo" or "test" script of the mode

bool TStarter::ApplyMode(const string &mode_name, const string &prev_mode){

    printer.LineBegin("Applying '" + mode_name + "' mode "
        "of the " + KeyWord() + " unit");

    if(mode_name == RESET_MODE){
        if(prev_mode.empty() || !Valid(prev_mode)){
            const string msg = prev_mode.empty() ?
                ("Previous " + KeyWord() + " mode not stated") :
                ("Stated previous " + KeyWord() + " mode " +
                    "'" + prev_mode + "' is not available");
            printer.LineAddWarn(msg);
            return true;
            }
        return RunScript(prev_mode, "undo");
        }

    return RunScript(mode_name, "do");
    }

//-----------------------------------------------------------------------------
// Check mode compliance

bool TStarter::CheckModeCompliance(const string &mode_name){

    printer.LineBegin("Checking '" + mode_name + "' " + KeyWord() +
        " mode compliance");

    return RunScript(mode_name, "test");
    }

//-----------------------------------------------------------------------------
