/*
    control++ : System configuration tool
    Copyright (C) 2017-2024 Alexey Appolonov

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "ulimits.h"
#include <sstream>

//-----------------------------------------------------------------------------
// Generate name for the file so it would be read last

// "...After reading /etc/security/limits.conf,
// individual files from the /etc/security/limits.d/ directory are read.
// The files are parsed one after another in the order of "C" locale.
// So the order will be special characters, numbers in ascending order,
// uppercase letters and lowercase letters in alphabetical order.
// If two files have same entry, then the entry read last
// will be taken in effect. Only files with *.conf extension
// will be read from this directory..."

string TULimits::GenerateFilename(void){

    // Define the active ulimis file (with the max name) among files of a
    // specified directory
    auto f = [](const string &dir_path, DIR *dir, string &result){
        string max_filename = SECURITY_LIM;
        string cur_filename;
        // For all files of the dir
        dirent *ent;
        while((ent = readdir(dir))){
            cur_filename = ent->d_name;
            if(CheckExtension(cur_filename, ".conf") &&
                    cur_filename > max_filename){
                max_filename = ent->d_name;
                }
            }
        max_filename = CutExtension(max_filename);
        result = DirEnding(dir_path);
        if(max_filename > SECURITY_LIM){
            string buf;
            for(const auto &c : max_filename){
                if(c < 'z'){
                    buf += c + 1;
                    break;
                    }
                else{
                    buf += c;
                    }
                }
            result += buf + "-" SECURITY_LIM;
            }
        else{
            result += SECURITY_LIM;
            }
        return true;
        };

    string active_conf;
    string err;

    // Opening the directory that stores ulimits files and applying the lambda
    if(!OpenAndPerform(SECURITY_DIR, f, active_conf, err)){
        printer.LineAddErr(err);
        }

    return active_conf;
    }

//-----------------------------------------------------------------------------
// Remove control++ limits from the "limits.d" dir

bool TULimits::CleanLimitsD(void){

    // Remove old configuration files from specified open dir
    auto f = [](const string &f_path, dirent *f_ent){
        if(f_ent->d_type == DT_REG){
            const int pos = f_path.rfind("control++.conf");
            if(pos >= 0){
                // Zero value on success
                if(remove(f_path.c_str())){
                    printer.LineAddErr("Can't remove old ulimits file "
                        "'" + f_path + "'");
                    return false;
                    }
                }
            }
        return true;
        };

    // Opening the directory that stores ulimits files and applying the lambda
    string err;
    if(!OpenAndPerformOnAllFiles(SECURITY_DIR, f, err)){
        printer.LineAddErr(err);
        return false;
        }

    return true;
    }

//-----------------------------------------------------------------------------
// Get a path of an active conf file

string TULimits::ActiveConfPath(void){

    // Lambda that will be evaluated in the contex of opened limits.d dir
    auto f = [](const string &dir_path, DIR *dir, string &result){
        const string right_part(SECURITY_LIM);
        // For all files of the dir
        dirent *ent;
        while((ent = readdir(dir))){
            if(ent->d_type != DT_REG){
                continue;
                }
            const string cur_filename = ent->d_name;
            const auto pos = cur_filename.rfind("control++.conf");
            if(pos == (cur_filename.size() - right_part.size()) &&
                    result < cur_filename){
                result = cur_filename;
                }
            }
        if(result.empty()){
            return false;
            }
        result = DirEnding(dir_path) + result;
        return true;
        };

    string active_conf;
    string err;

    // Opening the directory that stores ulimits files and applying the lambda
    if(!OpenAndPerform(SECURITY_DIR, f, active_conf, err)){
        printer.LineAddErr(err);
        }

    return active_conf;
    }

//-----------------------------------------------------------------------------
// Save current ulimits file in ULIMS_DIR dir to be able to restore it later

bool TULimits::SaveStateHeldBeforeModeSetting(void){

    const string active_conf = ActiveConfPath();
    const string dest = EnvPath(STATE_DIR) + KeyWord() + "/" RESET_MODE +
        (run_mode.Flag(M_MICRO_MODESETTING) ? "_micro" : "");

    if(!active_conf.empty() && !CopyFile(active_conf, dest)){
        printer.LineAddErr("Can't copy ulimits file "
            "'" + active_conf + "' to '" + dest + "'");
        return false;
        }

    return true;
    }

//-----------------------------------------------------------------------------
// Check mode compliance

bool TULimits::CheckModeCompliance(const string &mode_name){

    bool res = false;
    string status;
    string active_conf_path;
    string active_conf_data;
    string mode_desc_path;
    string mode_desc_data;

    // Read ulimits file to mem, trim spaces and comments and return the data
    auto GetULimitsData = [](const string &path){
        TFileInMem file;
        string data;
        if(file.ReadFileToMem(path)){
            // Put data string of a list-file into string stream
            istringstream iss(file.data);
            string line;
            // For all lines of the string stream
            while(getline(iss,line)){
                CutComment(line);
                Trim(line);
                // Replace each sequence of tabs and spaces with a single space
                // character, since delimiters between columns of ulimits files
                // (/etc/security/limits.d/*) can consist of any sequence of
                // tabs and spaces
                string line_thin;
                char prev_c = '0';
                for(char c : line){
                    if(!IsTabOrSpace(c) || !IsTabOrSpace(prev_c)){
                        if(IsTabOrSpace(c)){
                            c = ' ';
                            }
                        line_thin += c;
                        }
                    prev_c = c;
                    }
                if(!line_thin.empty()){
                    data += line_thin + "\n";
                    }
                }
            }
        return data;
        };

    printer.LineBegin("Checking '" + mode_name + "' " + KeyWord() + " "
        "mode compliance");

    // Validation of the given mode
    if(!Valid(mode_name)){
        status = "[ERROR: Mode is not available]";
        goto CheckModeComplianceReturn;
        }

    // Getting path of the active ulimits file
    active_conf_path = ActiveConfPath();
    if(active_conf_path.empty()){
        status = "[NEGATIVE: There is no active control++ ulimits file]";
        goto CheckModeComplianceReturn;
        }

    // Getting path of ulimits file of the mode to be checked
    mode_desc_path = ModeDescriptionPath(mode_name, NOT_A_DIR);

    // Getting and comparing data of the two ulimits files
    active_conf_data = GetULimitsData(active_conf_path);
    mode_desc_data = GetULimitsData(mode_desc_path);

    res = active_conf_data == mode_desc_data;

    status = res ?
        "[POSITIVE]" :
        "[NEGATIVE: Current active ulimits file is different]";

    CheckModeComplianceReturn:

    printer.LineEnd(status);

    return res;
    }

//-----------------------------------------------------------------------------
// Copy an ulimits file of a current mode to a dir where all ulimits files
// are stored

bool TULimits::ApplyMode(const string &mode_name, const string &prev_mode){

    // Determining a path of an ulimits file to be copied
    const string src = ModeDescriptionPath(mode_name, NOT_A_DIR);
    if(src.empty()){
        if(mode_name != RESET_MODE){
            printer.LineAddErr("Can't evaluate path of the ulimits file");
            return false;
            }
        // It's OK not to have 'reset' file
        printer.LineAddWarn("There is no 'reset' file");
        return true;
        }

    // Saving currently active ulimits file and
    // cleaning limits.d dir from all the control++ limits
    if((mode_name != RESET_MODE && !SaveStateHeldBeforeModeSetting()) ||
            !CleanLimitsD()){
        return false;
        }

    // Writing the new one
    const string dest = GenerateFilename();
    if(!CopyFile(src, dest)){
        printer.LineAddErr("Can't copy ulimits file "
            "'" + src + "' to '" + dest + "'");
        return false;
        }

    return true;
    }

//-----------------------------------------------------------------------------
