/*
    control++ : System configuration tool
    Copyright (C) 2017-2024 Alexey Appolonov

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "unit.h"
#include <sys/stat.h>

array<string, 3> available_unit_names = {KW_ULIMITS, KW_PERMS, KW_SCRIPTS};

//-----------------------------------------------------------------------------
// Mode name access

TUnitMode* TUnit::operator[](const int i){

    if(i > 0 && i < (int)available_modes.size()){
        return &(available_modes[i]);
        }

    return NULL;
    }

//-----------------------------------------------------------------------------
// Display list of available modes and point out a current mode

#define TAB "    "

void TUnit::PrintAvailableModes(
        const map<string, map<string, string>> &names_of_available_submodes){

    printer.LineEnd("Available " + keyword + " modes:");

    for(const auto &m : available_modes){

        const auto &m_name = m.Name();

        if(m_name.empty() || m_name == RESET_MODE){
            continue;
            }

        string line = m_name;
        vector<string> details;

        if(m.Current(MACRO)){
            details.push_back("current");
            }

        if(m.Current(MICRO)){
            details.push_back("current micro");
            }

        if(!details.empty()){
            line += " [";
            for(const auto &d : details){
                line += d + ", ";
                }
            line.pop_back();
            line.back() = ']';
            }

        printer.LineEnd(TAB + line);

        if(keyword != "macro"){
            continue;
            }

        // Display sub modes of every macro mode
        const auto &modes_of_units_it = names_of_available_submodes.find(m_name);
        if(modes_of_units_it == names_of_available_submodes.end()){
            continue;
            }
        const auto &[_, modes_of_units] = *modes_of_units_it;

        for(const string &unit_name : available_unit_names){
            const auto &submode_it = modes_of_units.find(unit_name);
            if(submode_it == modes_of_units.end()){
                continue;
                }
            const auto &[_, submode] = *submode_it;
            printer.LineEnd(TAB TAB + unit_name + ": " +
                (submode.empty() ? "-" : submode));
            }
        }
    }

#undef TAB

//-----------------------------------------------------------------------------
// Display current mode

void TUnit::PrintCurrentMode(const bool micro, const bool shrt){

    const string current_mode = CurrentModeName(micro);

    if(!micro && current_mode_stated_in_iconf != current_mode){
        string msg = "Mode description is changed since last mode setting - ";
        msg += current_mode_stated_in_iconf.empty() ?
            keyword + " was not set" :
            "'" + current_mode_stated_in_iconf + "' " + keyword + " was set";
        msg += ", but '" + current_mode + "' " + keyword + " is stated "
            "in the configuration file";
        printer.Warn(msg);
        }

    else if(!current_mode.empty()){
        if(current_mode == UNKNOWN_MODE){
            printer.Warn("Current " + keyword + (micro ? " micro" : "") +
                + " mode is unknown");
            }
        else{
            printer.LineEnd((shrt ? "" : "Current ") + keyword +
                (micro ? " micro" : "") + (shrt ? "" : " mode") + ": "
                "'" + current_mode + "'");
            }
        }

    else if(!micro){
        printer.LineEnd((shrt ? "" : "Current ") + keyword +
            (shrt ? "" : " mode") + ": (not specified)");
        }
    }

//-----------------------------------------------------------------------------
// Print DOM of all the config files

void TUnit::PrintConfDOM(void){

    for(const auto &conf : configs){
        conf.PrintDOM();
        }
    }

//-----------------------------------------------------------------------------
// Check if given name of the mode is a name of the special mode

bool TUnit::IsSpecialMode(const string &mode_name){

    return (mode_name == RESET_MODE || mode_name.empty());
    }

//-----------------------------------------------------------------------------
// Form vector of names of available modes

bool TUnit::FindoutAvailableModes(void){

    vector<string> modes;
    DIR *dir;
    const string dir_name = DirOfTheUnit();

    // Openning the directory of the unit
    if(!(dir = opendir(dir_name.c_str()))){
        // Could not open the dir
        printer.LineAddErr("Could not open '" + dir_name + "' dir");
        return false;
        }

    // For all files of the dir
    dirent *ent;
    while((ent = readdir(dir))){
        FindoutAvailableModesSpecialPart(ent, modes);
        }

    // Closing the dir and writing available modes
    closedir(dir);
    WriteAvailableModes(modes);

    return true;
    }

//-----------------------------------------------------------------------------
// Common realisation of the special part

void TUnit::FindoutAvailableModesSpecialPart(dirent *ent, vector<string> &modes){

    // Processing only regular files
    if(ent->d_type != DT_REG){
        return;
        }

    modes.push_back(ent->d_name);

    // Name of the file should not start with dot
    if(modes.back().empty() || modes.back().front() == '.'){
        modes.pop_back();
        }
    }

//-----------------------------------------------------------------------------
// Form full file path to the directory of the unit

string TUnit::DirOfTheUnit(void){

    return string(EnvPath(CONF_DIR) + keyword + "/");
    }

//-----------------------------------------------------------------------------
// Form full file path to a dir where prev state of the unit is stored

string TUnit::StateDirOfTheUnit(void){

    return string(EnvPath(STATE_DIR) + keyword + "/");
    }

//-----------------------------------------------------------------------------
// Form a full file path of a description of a specified mode and check whether
// this file does exist or doesn't (return an empty string if it doesn't exist)

string TUnit::ModeDescriptionPath(const string &mode_name, const int is_dir,
        const bool do_check){

    if(mode_name.empty()){
        return "";
        }

    string path;

    if(mode_name != RESET_MODE){
        path = DirOfTheUnit() + mode_name;
        }
    else{
        path = StateDirOfTheUnit() + mode_name;
        if(run_mode.Flag(M_MICRO_MODESETTING)){
            path += "_micro";
            }
        }

    if(do_check){
        if(!CheckFile(path)){
            return "";
            }
        // Add slash if the file is a dir (return an empty string if can't stat
        // the file)
        struct stat path_stat;
        if(stat(path.c_str(), &path_stat) == -1){
            return "";
            }
        if(S_ISDIR(path_stat.st_mode)){
            if(is_dir != IS_DIR){
                return "";
                }
            path += "/";
            }
        }
    else if(is_dir == IS_DIR){
        path += "/";
        }

    return path;
    }

//-----------------------------------------------------------------------------
// Write names of modes to the "modes" vector

// There are two special modes - "reset" and empty (or "ghost") mode,
// Setting "reset" mode means calling special method of the units,
// setting empty mode means "don't do anything"

void TUnit::WriteAvailableModes(const vector<string> &modes,
        const map<string, map<string, string>> &submodes){

    available_modes.clear();

    for(auto &m : modes){
        available_modes.push_back(TUnitMode(m));
        const auto &submodes_of_the_mode = submodes.find(m);
        if(submodes_of_the_mode != submodes.end()){
            available_modes.back().WriteSubModes(submodes_of_the_mode->second);
            }
        }

    available_modes.push_back(TUnitMode(RESET_MODE));
    available_modes.push_back(TUnitMode(""));
    }

//-----------------------------------------------------------------------------
// Write current mode

bool TUnit::WriteCurrentMode(const string &mode_name, const bool micro){

    TUnitMode *current_mode = CurrentMode();
    TUnitMode *new_mode = (mode_name == RESET_MODE) ?
        FindMode("") :
        FindMode(mode_name);

    // Switch prev current mode off and new mode on
    if(new_mode){
        if(current_mode){
            current_mode->Switch(false, micro);
            }
        new_mode->Switch(true, micro);
        return true;
        }

    return false;
    }

//-----------------------------------------------------------------------------
// Get pointer to a current mode of the unit

TUnitMode* TUnit::CurrentMode(const bool micro){

    for(auto &m : available_modes){
        if(m.Current(micro)){
            return &m;
            }
        }

    return NULL;
    }

//-----------------------------------------------------------------------------
// Get name of a current mode

string TUnit::CurrentModeName(const bool micro){

    TUnitMode* current = CurrentMode(micro);
    if(!current){
        return UNKNOWN_MODE;
        }

    return current->Name();
    }

//-----------------------------------------------------------------------------
// Find mode with a given name

TUnitMode* TUnit::FindMode(const string &new_mode){

    for(auto &m : available_modes){
        if(m.Name() == new_mode){
            return &m;
            }
        }

    return NULL;
    }

//-----------------------------------------------------------------------------
// Check if given mode is available

bool TUnit::Valid(const string &mode_name){

    if(mode_name.empty()){
        return true;
        }

    return FindMode(mode_name) ? true : false;
    }

//-----------------------------------------------------------------------------
// Set given mode

bool TUnit::SetMode(const string &mode_name, const string &prev_mode){

    if(Valid(mode_name)){
        configs.clear();
        if(mode_name.empty() || ApplyMode(mode_name, prev_mode)){
            return WriteCurrentMode(mode_name);
            }
        }
    else{
        printer.LineAddErr("Unknown " + keyword);
        }

    return false;
    }

//-----------------------------------------------------------------------------
// Write full path of a new configuration file

bool TUnit::NewConfFile(const string &conf_path_or_mode_name, const int mark){

    const string conf_path = (mark < 0) ?
        FormatFilePath(conf_path_or_mode_name) :
        ModeDescriptionPath(conf_path_or_mode_name, mark, false);

    // Checking existence of the conf file
    if(conf_path.empty() || !CheckFile(conf_path)){
        bool no_reset_data = false;
        if(conf_path_or_mode_name == RESET_MODE){
            if(!CurrentModeName().empty()){
                no_reset_data = true;
                printer.LineEnd("[INTERRUPTION]");
                cout << "It seems that mode was set, but there is no data "
                        "to reset the mode" << endl;
                if(run_mode.Force()){
                    printer.LineEnd("Acknowledging the '--force' flag, "
                        "the operation will be continued");
                    }
                else if(!Approval("Continue anyway [y/N]? ", "y")){
                    return false;
                    }
                printer.LineRestart();
                }
            }
        // It's OK if reset/internal conf file is not exist yet
        if(no_reset_data || conf_path == EnvPath(INTERNAL_CONF_FILE)){
            FILE *file = fopen(conf_path.c_str(), "a+");
            if(!file){
                printer.LineAddErr("Can't create config file "
                    "'" + conf_path + "'");
                return false;
                }
            fclose(file);
            }
        else{
            printer.LineAddErr("Can't find config file for the " + KeyWord());
            return false;
            }
        }

    TConf conf;
    conf.WriteFilePath(conf_path);
    configs.push_back(conf);

    return true;
    }

//-----------------------------------------------------------------------------
// Configure the unit

bool TUnit::Configure(void){

    const auto err = [](const TConf &conf){
        const string e = conf.Err();
        return e.empty() ? "" : " (" + e + ")";
        };

    // Going over all config files
    for(auto &conf : configs){
        if(conf.FilePath().empty()){
            printer.LineAddErr("There is no configuration file path");
            return false;
            }
        // Reading configuration ini-file to memory
        if(!conf.ReadFile()){
            printer.LineAddErr(
                "Can't read " + KeyWord() + " configuration file" + err(conf)
                );
            return false;
            }
        // Parsing ini-file (forming DOM)
        if(!conf.Parse()){
            printer.LineAddErr(
                "Wrong " + KeyWord() + " configuration file format" + err(conf)
                );
            return false;
            }
        }

    // Applying DOM of the configuration files
    if(!ApplyConfDOM()){
        if(!run_mode.Flag(M_SILENT_TERMINATION)){
            printer.LineAddErr("Can't apply given configuration "
                "for the " + KeyWord());
            }
        return false;
        }

    // Unit is configured successfully
    return true;
    }

//-----------------------------------------------------------------------------
