/*
    control++ : System configuration tool
    Copyright (C) 2017-2024 Alexey Appolonov

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef unitH
#define unitH

#include "main.h"

#ifdef DEBUGING
#include "conf.h"
#include "vector.h"
#else
#include <libcontrol++/conf.h>
#include <libcontrol++/vector.h>
#endif

#include <map>

#define MICRO     true
#define MACRO     false
#define SHORT_FMT true

#define KW_ULIMITS  "ulimits"     // Keyword of the ulimits control unit
#define KW_PERMS    "permissions" // Keyword of the permissions control unit
#define KW_SCRIPTS  "scripts"     // Keyword of the script starter control unit

//-----------------------------------------------------------------------------
// Mode of the unit

class TUnitMode{
    private:
        bool current = false;       // Flag is true if mode is current
                                    //  for the unit
        bool current_micro = false; // Flag is true if mode is current
                                    //  micro mode for the unit
        string name;                // Name of the mode
        map<string, string> submodes;   // Key - name of the cotrolled unit,
                                        // value - name of the submode
    public:
        // Fields access
        const string& Name(void) const{return name;}
        const map<string, string>& SubModes(void) const{return submodes;}
        bool Current(bool micro=false) const{
            if(micro){
                return current_micro;
                }
            return current;
            }
        // Modifiers
        void Switch(bool on, bool micro=false){
            if(micro){
                current_micro = on;
                }
            else{
                current = on;
                }
            }
        void WriteName(const string& n){name = n;}
        void WriteSubModes(const map<string, string>& sm){submodes = sm;}

        TUnitMode(const string &n) : name(n){}
    };

//-----------------------------------------------------------------------------
// Control unit

class TUnit{

    protected:

        string keyword;                     // Keyword for output info
        vector<TConf> configs;              // Configurations
        vector<TUnitMode> available_modes;  // Available modes of the unit
        string current_mode_stated_in_iconf;

        // Find current mode of the unit
        TUnitMode* CurrentMode(bool micro=false);
        // Find mode with a given name
        TUnitMode* FindMode(const string &mode_name);
        // Apply current mode with optional param
        virtual bool ApplyMode(const string &mode_name,
                const string &prev_mode="")=0;
        // Apply configuration that stored in the configuration file DOM
        virtual bool ApplyConfDOM(void){return true;}
        // Form full file path to the directory of the unit
        string DirOfTheUnit(void);
        // Form full file path to a dir where prev state of the unit is stored
        string StateDirOfTheUnit(void);
        // Form full file path and check whether it exists or not
        string ModeDescriptionPath(const string &mode_name, const int is_dir,
                const bool do_check=true);
        // Special part of the FindoutAvailableModes function
        virtual void FindoutAvailableModesSpecialPart(dirent *ent,
                vector<string> &modes);

    public:

        // Mode name access
        TUnitMode* operator[](const int i);
        // Get name of a current mode
        string CurrentModeName(const bool micro=false);
        string CurrentModeStatedInIConf(void){
            return current_mode_stated_in_iconf;
            }
        // Get unit's keyword
        const string& KeyWord(void){
            return keyword;
            }
        // Display list of available modes and point out a current mode
        void PrintAvailableModes(
            const map<string, map<string, string>> &names_of_available_submodes={});
        // Display current mode
        void PrintCurrentMode(const bool micro=false, const bool shrt=false);
        // Display configuration file DOM
        void PrintConfDOM(void);
        // Check if given mode is available
        bool Valid(const string &mode_name);
        // Check if given name of the mode is a name of the special mode
        bool IsSpecialMode(const string &mode_name);
        // Write full path of a new configuration file
        bool NewConfFile(const string &conf_path, const int mark=-1);
        // Configure the unit
        virtual bool Configure(void);
        // Form vector of names of available modes
        virtual bool FindoutAvailableModes(void);
        // Write names of modes to the "modes" vector
        void WriteAvailableModes(const vector<string> &modes,
                const map<string, map<string, string>>
                    &submodes=map<string, map<string, string>>());
        // Write current mode
        bool WriteCurrentMode(const string &mode_name, const bool micro=false);
        void WriteCurrentModeStatedInIConf(const string &mode_name){
            current_mode_stated_in_iconf=mode_name;
            }
        // Set given mode
        virtual bool SetMode(const string &mode_name,
                const string &prev_mode="");
        // Check mode compliance
        virtual bool CheckModeCompliance(const string &mode_name)=0;

        virtual ~TUnit(void){};
    };

//-----------------------------------------------------------------------------

#endif
